/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.kanji;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;

import org.vono.narau.Common;

import android.database.Cursor;
import android.database.SQLException;
import android.database.sqlite.SQLiteDatabase;
import android.util.Log;

/**
 * 
 * Read the kanji svg database.
 * 
 */
public class KanjiDB {
	private static final String TAG = KanjiDB.class.getSimpleName();

	public static final int VALUE_DB_VERSION = 2;

	private static final String TABLE_RADPOS = "radpos"; // rapical + position
	private static final String TABLE_KANJI = "kanji";
	private static final String TABLE_MEANINGS = "meanings";
	/**
	 * can't used 'group' as the table name, it is a reserved word, the produced
	 * db can't be read.
	 */
	private static final String TABLE_GROUP = "kgroup";
	private static final String TABLE_PATH = "path";
	private static final String VIEW_SVG = "kanjivg";

	private static final String WHERE_CODEPOINT = "codepoint=?";

	private static final String COL_KANJI_CODEPOINT = "codepoint";
	private static final String COL_KANJI_ELEMENT = "element";
	private static final String COL_KANJI_RADICAL = "radical";
	private static final String COL_KANJI_GRADE = "grade";
	private static final String COL_KANJI_FREQUENCY = "frequency";
	private static final String COL_KANJI_JLPT = "jlpt";
	private static final String COL_KANJI_STROKE_COUNT = "stroke_count";
	private static final String COL_KANJI_READING_ON = "reading_on";
	private static final String COL_KANJI_READING_KUN = "reading_kun";

	private static final String[] COLS_KANJI_WITH_READING = { COL_KANJI_CODEPOINT,
			COL_KANJI_ELEMENT, COL_KANJI_RADICAL, COL_KANJI_GRADE, COL_KANJI_FREQUENCY,
			COL_KANJI_JLPT, COL_KANJI_STROKE_COUNT, COL_KANJI_READING_ON, COL_KANJI_READING_KUN };
	private static final String[] COLS_KANJI_NO_READING = { COL_KANJI_CODEPOINT, COL_KANJI_ELEMENT,
			COL_KANJI_RADICAL, COL_KANJI_GRADE, COL_KANJI_FREQUENCY, COL_KANJI_JLPT,
			COL_KANJI_STROKE_COUNT };
	private static final String[] COLS_KANJIVG_CODEPOINT = { COL_KANJI_CODEPOINT };
	private static final String[] COLS_RADPOS = { "id", "value" };
	private static final String[] COLS_MEANINGS = { "lang", "meanings" };
	private static final String[] COLS_GROUPS = { "id", "parent", "element", "position", "radical" };
	private static final String[] COLS_PATH = { "parent", "type", "draw" };
	private static final String ORDER_BY_CODEPOINT = "codepoint asc";

	public static final String KANJI_WHERE_CODEPOINT = "codepoint=?";

	private static String dbFullName = null;
	private static boolean goodVersion;

	private static String[] radicals = null;
	private static String[] positions = null;

	/**
	 * 
	 * @return true if the db exists and have a good format version.
	 */
	public static boolean initDB() {
		KanjiDB.dbFullName = Common.dataPath + File.separator + Common.DB_KANJI_NAME;
		KanjiDB.goodVersion = false;

		// check db version
		SQLiteDatabase db = null;
		Cursor cursor = null;
		File dbFile = new File(KanjiDB.dbFullName);
		if (dbFile.canRead()) {
			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				cursor = db.query(Common.TABLE_DBINFO, Common.COLS_DBINFO, null, null, null, null,
						null);
				if (cursor.moveToFirst()) {
					int idxColKey = cursor.getColumnIndex(Common.COLS_DBINFO[0]);
					int idxColValueInt = cursor.getColumnIndex(Common.COLS_DBINFO[2]);
					do {
						String key = cursor.getString(idxColKey);
						if (Common.KEY_DB_VERSION.equals(key)) {
							int version = cursor.getInt(idxColValueInt);
							if (VALUE_DB_VERSION == version) {
								KanjiDB.goodVersion = true;
							}
						}
					} while (cursor.moveToNext());
				}
				cursor.close();

				if (KanjiDB.goodVersion) {
					cursor = db.query(TABLE_RADPOS, COLS_RADPOS, null, null, null, null, null);
					if (cursor.moveToFirst()) {
						int colIdxId = cursor.getColumnIndex(COLS_RADPOS[0]);
						int colIdxVal = cursor.getColumnIndex(COLS_RADPOS[1]);
						int count = cursor.getCount();

						String[] rads = new String[count];
						String[] pos = new String[count];
						int numRad = 0;
						int numPos = 0;

						// FIXME fragile code
						do {
							String idString = cursor.getString(colIdxId);
							String value = cursor.getString(colIdxVal);
							int id = Integer.parseInt(idString.substring(1));

							if (idString.startsWith("r")) {
								rads[id] = value;
								numRad = id > numRad ? id : numRad;

							} else if (idString.startsWith("p")) {
								pos[id] = value;
								numPos = id > numPos ? id : numPos;
							}
						} while (cursor.moveToNext());

						KanjiDB.radicals = Arrays.copyOf(rads, numRad + 1);
						KanjiDB.positions = Arrays.copyOf(pos, numPos + 1);

					}
				}

			} catch (SQLException se) {

			} finally {
				if (null != cursor) {
					cursor.close();
				}
				if (null != db) {
					db.close();
				}
			}

		}
		return KanjiDB.goodVersion;
	}

	public static void clean() {
		KanjiDB.goodVersion = false;
		KanjiDB.dbFullName = null;
		KanjiDB.radicals = null;
		KanjiDB.positions = null;
	}

	public static boolean isReady() {
		return KanjiDB.goodVersion;
	}

	public static ArrayList<String> getLangs() {
		ArrayList<String> langs = null;
		if (KanjiDB.goodVersion) {
			SQLiteDatabase db = null;
			Cursor cursor = null;
			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				cursor = db.query(true, TABLE_MEANINGS, new String[] { "lang" }, null, null, null,
						null, null, null);

				if (cursor.moveToFirst()) {
					langs = new ArrayList<String>(cursor.getCount());

					do {
						Common.addLangToList(cursor.getString(0), langs);
					} while (cursor.moveToNext());
				}

			} finally {
				if (null != cursor) {
					cursor.close();
				}

				if (null != db) {
					db.close();
				}
			}
		}

		return langs;
	}

	/**
	 * 
	 * @param codepoint
	 *            unicode value of the kanji
	 * @return The loaded kanji (with meaning and svg path) or null if not found
	 */
	public static Kanji getKanji(int codepoint) {
		Kanji kanji = null;

		if (KanjiDB.goodVersion) {
			SQLiteDatabase db = null;
			Cursor cursor = null;
			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);

				String[] selectionArgs = new String[1];
				selectionArgs[0] = Integer.toString(codepoint);
				cursor = db.query(TABLE_KANJI, COLS_KANJI_WITH_READING, KANJI_WHERE_CODEPOINT,
						selectionArgs, null, null, null);
				if (cursor.moveToFirst()) {
					kanji = KanjiDB.cursorToKanji(db, cursor, true, true, true);
				}
			} finally {
				if (null != cursor)
					cursor.close();
				if (null != db)
					db.close();
			}
		}
		return kanji;
	}

	public static void getKanjiVG(Kanji kanji) {
		if (KanjiDB.goodVersion) {
			SQLiteDatabase db = null;
			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);
				KanjiDB.getKanjiVG(kanji, db);
			} finally {
				if (null != db)
					db.close();
			}
		}
	}

	public static void getMeaning(Kanji kanji) {
		if (KanjiDB.goodVersion) {
			SQLiteDatabase db = null;
			try {
				db = SQLiteDatabase.openDatabase(dbFullName, null, SQLiteDatabase.OPEN_READONLY
						| SQLiteDatabase.NO_LOCALIZED_COLLATORS);
				KanjiDB.getMeanings(kanji, db);
			} finally {
				if (null != db)
					db.close();
			}
		}
	}

	/**
	 * 
	 * @param where
	 *            The query where clause.
	 * @return An iterator for the kanjis
	 */
	public static KanjiDBIterator getKanjiIterator(String selection, String[] selectionArgs,
			boolean loadMeaning, boolean loadSvg, boolean loadReading) {
		if (!KanjiDB.goodVersion) {
			return null;
		}

		SQLiteDatabase db = SQLiteDatabase.openDatabase(dbFullName, null,
				SQLiteDatabase.OPEN_READONLY | SQLiteDatabase.NO_LOCALIZED_COLLATORS);

		Cursor cursor = null;

		if (loadReading) {
			cursor = db.query(TABLE_KANJI, COLS_KANJI_WITH_READING, selection, selectionArgs, null,
					null, ORDER_BY_CODEPOINT);
		} else {
			cursor = db.query(TABLE_KANJI, COLS_KANJI_NO_READING, selection, selectionArgs, null,
					null, ORDER_BY_CODEPOINT);
		}

		KanjiDB.KanjiDBIterator ite = new KanjiDB.KanjiDBIterator(db, cursor, loadMeaning, loadSvg,
				loadReading);
		return ite;
	}

	/**
	 * Return an iterator on kanjis codepoint where an SVG drawing is available.
	 * 
	 * @param selection
	 *            where clause
	 * @param selectionArgs
	 *            where values
	 * @return an KanjiSVGIterator or null if the database do not have the good
	 *         version.
	 */
	public static KanjiSVGIterator getKanjiSVGIterator(String selection, String[] selectionArgs) {
		if (!KanjiDB.goodVersion) {
			return null;
		}
		// Date start = new Date();

		SQLiteDatabase db = SQLiteDatabase.openDatabase(dbFullName, null,
				SQLiteDatabase.OPEN_READONLY | SQLiteDatabase.NO_LOCALIZED_COLLATORS);

		Cursor cursor = db.query(VIEW_SVG, COLS_KANJIVG_CODEPOINT, selection, selectionArgs, null,
				null, ORDER_BY_CODEPOINT);

		KanjiDB.KanjiSVGIterator ite = new KanjiSVGIterator(db, cursor);

		// Date end = new Date();
		// long time = end.getTime() - start.getTime();
		// FIXME Perf: takes more than 1s to get the iterator
		// Log.i(TAG, "getKanjiSVGIterator: " + (time / 1000.0f) + "s");

		return ite;
	}

	private static void getMeanings(Kanji kanji, SQLiteDatabase db) {

		String[] selectionArgs = new String[1];
		selectionArgs[0] = Integer.toString(kanji.codepoint);
		Cursor cursor = db.query(TABLE_MEANINGS, COLS_MEANINGS, WHERE_CODEPOINT, selectionArgs,
				null, null, null);

		if (cursor.moveToFirst()) {
			int idxColLang = cursor.getColumnIndex(COLS_MEANINGS[0]);
			int idxColMeaning = cursor.getColumnIndex(COLS_MEANINGS[1]);

			HashMap<String, String> map = null;
			if (null == kanji.meanings) {
				kanji.meanings = new HashMap<String, String>(1);
			}
			map = kanji.meanings;

			do {

				String lang = cursor.getString(idxColLang);
				String meanings = cursor.getString(idxColMeaning);
				map.put(lang, meanings);

			} while (cursor.moveToNext());
		}

		cursor.close();
	}

	private static void getKanjiVG(Kanji kanji, SQLiteDatabase db) {
		// Group
		// id INTEGER, parent INTEGER, element TEXT, position INTEGER,
		// radical INTEGER

		Kanji.Group[] groups = null;
		String[] selectionArgs = new String[1];
		selectionArgs[0] = Integer.toString(kanji.codepoint);
		Cursor cursor = db.query(TABLE_GROUP, COLS_GROUPS, WHERE_CODEPOINT, selectionArgs, null,
				null, "id ASC");

		if (cursor.moveToFirst()) {
			int idxColId = cursor.getColumnIndex(COLS_GROUPS[0]);
			int idxColParent = cursor.getColumnIndex(COLS_GROUPS[1]);
			int idxColElement = cursor.getColumnIndex(COLS_GROUPS[2]);
			int idxColPosition = cursor.getColumnIndex(COLS_GROUPS[3]);
			int idxColRadical = cursor.getColumnIndex(COLS_GROUPS[4]);

			if (null == kanji.group) {
				kanji.group = new Kanji.Group(0);
			}

			groups = new Kanji.Group[cursor.getCount() + 1];
			groups[0] = kanji.group;

			do {
				int id = cursor.getInt(idxColId);
				int parent = cursor.getInt(idxColParent);
				String element = cursor.getString(idxColElement);
				String position = cursor.isNull(idxColPosition) ? null : KanjiDB.positions[cursor
						.getInt(idxColPosition)];

				String radical = cursor.isNull(idxColRadical) ? null : KanjiDB.radicals[cursor
						.getInt(idxColRadical)];

				Kanji.Group group = new Kanji.Group(id);
				group.element = element;
				group.position = position;
				group.radical = radical;

				if (null == groups[parent].elements) {
					groups[parent].elements = new ArrayList<Kanji.Element>(1);
				}
				groups[parent].elements.add(group);
				groups[id] = group;

			} while (cursor.moveToNext());
		}

		cursor.close();

		// Path
		// codepoint INTEGER NOT NULL, parent INTEGER NOT NULL, type TEXT, draw
		// TEXT NOT NULL
		cursor = db.query(TABLE_PATH, COLS_PATH, WHERE_CODEPOINT, selectionArgs, null, null,
				"parent ASC");
		if (cursor.moveToFirst()) {
			int idxColParent = cursor.getColumnIndex(COLS_PATH[0]);
			int idxColType = cursor.getColumnIndex(COLS_PATH[1]);
			int idxColDraw = cursor.getColumnIndex(COLS_PATH[2]);

			if (null == kanji.group) {
				kanji.group = new Kanji.Group(0);
			}
			if (null == groups) {
				groups = new Kanji.Group[1];
				groups[0] = kanji.group;
			}
			int numGroups = groups.length;
			int id = 1;

			do {
				int parent = cursor.getInt(idxColParent);
				String type = cursor.getString(idxColType);
				String draw = cursor.getString(idxColDraw);

				if (parent < numGroups) {
					Kanji.Path path = new Kanji.Path(id, type, draw);
					if (null == groups[parent].elements) {
						groups[parent].elements = new ArrayList<Kanji.Element>(1);
					}
					groups[parent].elements.add(path);
				} else {
					Log.e(TAG, "Error in parenting path for kanji " + kanji.codepoint
							+ ", no parent group id " + parent);
				}

			} while (cursor.moveToNext());
		}
		cursor.close();
	}

	private static Kanji cursorToKanji(SQLiteDatabase db, Cursor cursor, boolean loadMeaning,
			boolean loadSVG, boolean loadReading) {
		if (cursor.isClosed() || cursor.isAfterLast() || cursor.isBeforeFirst())
			return null;

		int idxColCodepoint = cursor.getColumnIndex(COL_KANJI_CODEPOINT);
		int idxColElement = cursor.getColumnIndex(COL_KANJI_ELEMENT);
		int idxColRadical = cursor.getColumnIndex(COL_KANJI_RADICAL);
		int idxColGrade = cursor.getColumnIndex(COL_KANJI_GRADE);
		int idxColFrequency = cursor.getColumnIndex(COL_KANJI_FREQUENCY);
		int idxColJLPT = cursor.getColumnIndex(COL_KANJI_JLPT);
		int idxColStrokeCount = cursor.getColumnIndex(COL_KANJI_STROKE_COUNT);
		int idxColReadingOn = cursor.getColumnIndex(COL_KANJI_READING_ON);
		int idxColReadingKun = cursor.getColumnIndex(COL_KANJI_READING_KUN);

		int codepoint = cursor.getInt(idxColCodepoint);
		Kanji kanji = new Kanji(codepoint);

		boolean notNullElement = !cursor.isNull(idxColElement);
		boolean notNullRadical = !cursor.isNull(idxColRadical);
		if (notNullElement || notNullRadical) {
			if (null == kanji.group) {
				kanji.group = new Kanji.Group(0);
			}
			if (notNullElement)
				kanji.group.element = cursor.getString(idxColElement);
			if (notNullRadical)
				kanji.group.radical = KanjiDB.radicals[cursor.getInt(idxColRadical)];
		}

		if (!cursor.isNull(idxColGrade))
			kanji.grade = cursor.getInt(idxColGrade);

		if (!cursor.isNull(idxColFrequency))
			kanji.frequency = cursor.getInt(idxColFrequency);

		if (!cursor.isNull(idxColJLPT))
			kanji.jlpt = cursor.getInt(idxColJLPT);

		if (!cursor.isNull(idxColStrokeCount))
			kanji.strokeCount = cursor.getInt(idxColStrokeCount);

		if (loadReading && idxColReadingKun != -1 && idxColReadingOn != -1) {
			kanji.readingsOn = cursor.getString(idxColReadingOn);
			kanji.readingsKun = cursor.getString(idxColReadingKun);
		}

		if (loadMeaning)
			KanjiDB.getMeanings(kanji, db);
		if (loadSVG)
			KanjiDB.getKanjiVG(kanji, db);

		return kanji;
	}

	/**
	 * Iterator on a kanji db
	 * 
	 */
	public static class KanjiDBIterator implements Iterator<Kanji> {
		private final Cursor cursor;
		private final SQLiteDatabase db;
		private int current;
		private int count;
		private boolean loadMeaning;
		private boolean loadSVG;
		private boolean loadReading;

		private KanjiDBIterator(SQLiteDatabase db, Cursor cursor, boolean loadMeaning,
				boolean loadSVG, boolean loadReading) {
			this.db = db;
			this.cursor = cursor;
			this.current = 0;
			this.count = cursor.getCount();
			this.loadMeaning = loadMeaning;
			this.loadSVG = loadSVG;
			this.loadReading = loadReading;
		}

		/**
		 * 
		 * @return The current kanji number or -1 if this iterator is closed or
		 *         0 the first() method was not called.
		 */
		public int getCurrent() {
			return this.current;
		}

		/**
		 * 
		 * @return The number of resuls in the query.
		 */
		public int getCount() {
			return this.count;
		}

		/**
		 * 
		 * @return the first kanji, or null if there is no kanji in the list.
		 */
		public Kanji first() {
			Kanji kanji = null;
			if (this.cursor.moveToFirst()) {
				kanji = KanjiDB.cursorToKanji(db, cursor, this.loadMeaning, this.loadSVG,
						this.loadReading);
				this.current = 0;
			}
			return kanji;
		}

		/**
		 * 
		 * @return The next kanji.
		 */
		public Kanji next() {
			Kanji kanji = null;
			if (!this.cursor.isClosed() && this.cursor.moveToNext()) {
				kanji = KanjiDB.cursorToKanji(db, cursor, this.loadMeaning, this.loadSVG,
						this.loadReading);
				this.current++;
			}
			return kanji;
		}

		/**
		 * 
		 * @return The previous kanji
		 */
		public Kanji previous() {
			Kanji kanji = null;
			if (this.cursor.moveToPrevious()) {
				kanji = KanjiDB.cursorToKanji(db, cursor, this.loadMeaning, this.loadSVG,
						this.loadReading);
				this.current--;
			}
			return kanji;
		}

		/**
		 * 
		 * @return The last kanji
		 */
		public Kanji last() {
			Kanji kanji = null;
			if (this.cursor.moveToLast()) {
				kanji = KanjiDB.cursorToKanji(db, cursor, this.loadMeaning, this.loadSVG,
						this.loadReading);
				this.current = this.count;
			}
			return kanji;
		}

		/**
		 * Move the the iterator to the given position
		 * 
		 * @param newPosition
		 *            The new position (>=0 && < count)
		 * @return The kanji, or null if the cursor is close or the new position
		 *         is out of range.
		 */
		public Kanji position(int newPosition) {
			Kanji kanji = null;
			if (!this.cursor.isClosed() && newPosition >= 0 && newPosition < this.count) {
				if (this.cursor.moveToPosition(newPosition)) {
					this.current = newPosition;
					kanji = KanjiDB.cursorToKanji(db, cursor, this.loadMeaning, this.loadSVG,
							this.loadReading);
				}
			}
			return kanji;
		}

		/**
		 * Get the specific kanji at index position
		 * 
		 * @param index
		 *            The kanji position
		 * @return the kanji at index position or null if index is invalid.
		 */
		public Kanji get(int index) {
			Kanji kanji = null;
			if (this.cursor.moveToPosition(index)) {
				kanji = KanjiDB.cursorToKanji(db, cursor, this.loadMeaning, this.loadSVG,
						this.loadReading);
				this.cursor.moveToPosition(this.current);
			}
			return kanji;
		}

		/**
		 * Must be call, to close the db cursor and the database.
		 */
		public void close() {
			this.cursor.close();
			this.db.close();
			this.current = -1;
			this.count = 0;
		}

		public boolean hasNext() {
			return !this.cursor.isClosed() && !this.cursor.isLast();
		}

		public boolean hasPrevious() {
			return !this.cursor.isClosed() && !this.cursor.isFirst();
		}

		/**
		 * throw UnsupportedOperationException.
		 */
		public void remove() {
			throw new UnsupportedOperationException("Can't remove a Kanji from the database");

		}
	}

	/**
	 * Iterator on codepoint which have an SVG path available.
	 * 
	 */
	public static class KanjiSVGIterator implements Iterator<Integer> {
		private int codepoint;

		private final Cursor cursor;
		private final SQLiteDatabase db;
		private final int count;
		private int current;

		private KanjiSVGIterator(SQLiteDatabase db, Cursor cursor) {
			this.db = db;
			this.cursor = cursor;
			this.count = cursor.getCount();
			this.codepoint = -1;
			this.current = 0;
		}

		/**
		 * 
		 * @return The number of resuls in the query.
		 */
		public int getCount() {
			return this.count;
		}

		/**
		 * 
		 * @return The current kanji number or -1 if this iterator is closed or
		 *         0 the first() method was not called.
		 */
		public int getCurrent() {
			return this.current;
		}

		public boolean hasNext() {
			return !this.cursor.isClosed() && !this.cursor.isLast();
		}

		/**
		 * Get the next kanji codepoint.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer next() {
			this.codepoint = -1;
			if (!this.cursor.isAfterLast() && !this.cursor.isClosed()) {
				if (cursor.moveToNext()) {
					this.codepoint = cursor.getInt(0);
					this.current++;
				}
			}
			return this.codepoint;
		}

		/**
		 * Get the previous kanji codepoint.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer previous() {
			this.codepoint = -1;
			if (!this.cursor.isBeforeFirst() && !this.cursor.isClosed()) {
				if (cursor.moveToPrevious()) {
					this.codepoint = cursor.getInt(0);
					this.current--;
				}
			}
			return this.codepoint;
		}

		/**
		 * Get the first kanji.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer first() {
			this.codepoint = -1;
			if (!this.cursor.isClosed()) {
				if (cursor.moveToFirst()) {
					this.codepoint = cursor.getInt(0);
					this.current = 0;
				}
			}
			return this.codepoint;
		}

		/**
		 * Get the last kanji.
		 * 
		 * @return the unicode value or -1 if there is no next value.
		 */
		public Integer last() {
			this.codepoint = -1;
			if (!this.cursor.isClosed()) {
				if (cursor.moveToLast()) {
					this.codepoint = cursor.getInt(0);
					this.current = this.count;
				}
			}
			return this.codepoint;
		}

		/**
		 * Move the the iterator to the given position
		 * 
		 * @param newPosition
		 *            The new position (>=0 && < count)
		 * @return The kanji codepoint, or -1 if the cursor is close or the new
		 *         position is out of range.
		 */
		public Integer position(int newPosition) {
			this.codepoint = -1;
			if (!this.cursor.isClosed() && newPosition >= 0 && newPosition < this.count) {
				if (this.cursor.moveToPosition(newPosition)) {
					this.current = newPosition;
					this.codepoint = cursor.getInt(0);
				}
			}
			return this.codepoint;
		}

		/**
		 * Must be call, to close the db cursor and the database.
		 */
		public void close() {
			this.cursor.close();
			this.db.close();
			this.current = -1;
			this.codepoint = -1;
		}

		/**
		 * UnsupportedOperationException.
		 */
		public void remove() {
			throw new UnsupportedOperationException("Can't remove a Kanji from the database");

		}

	}
}
