/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.jmdict;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.HashMap;
import java.util.Set;

import org.tmatesoft.sqljet.core.SqlJetEncoding;
import org.tmatesoft.sqljet.core.SqlJetException;
import org.tmatesoft.sqljet.core.SqlJetTransactionMode;
import org.tmatesoft.sqljet.core.table.ISqlJetCursor;
import org.tmatesoft.sqljet.core.table.ISqlJetTable;
import org.tmatesoft.sqljet.core.table.SqlJetDb;
import org.vono.narau.Common;
import org.vono.narau.jmdict.entry.Entry;
import org.vono.narau.jmdict.entry.KanjiElement;
import org.vono.narau.jmdict.entry.ReadingElement;
import org.vono.narau.jmdict.entry.Sense;
import org.vono.narau.jmdict.entry.SenseGloss;
import org.xmlpull.v1.XmlPullParserException;

/**
 * 
 *
 */
public class MainJMDict {

	private static final int NUM_LANGS = 4;
	private static final int NUM_GLOSS_MAX = 177493;
	private static final int NUM_KANJI = 155422;
	private static final int NUM_READING = 177097;
	public static final String EMPTY_STRING = "";
	public static final String SLASH = "/";

	private static final String KEY_DB_VERSION = "version";
	private static final int VALUE_DB_VERSION = 1;
	private static final String KEY_DB_CREATION_DATE = "creation_date";
	private static final String VALUE_CREATION_DATE = String.format("%1$tY%1$tm%1$td",
			Calendar.getInstance());

	private static final String TABLE_DBINFO = "db_info";
	private static final String TABLE_DICTINFO = "dict_info";
	private static final String TABLE_KANJI_ELEM = "k_ele";
	private static final String TABLE_READING_ELEM = "r_ele";
	private static final String TABLE_SENSE = "sense";
	private static final String TABLE_GLOSS = "gloss_%1$s";

	private static final String CREATE_TABLE_DB_INFO = "CREATE TABLE " + TABLE_DBINFO
			+ " (key TEXT PRIMARY KEY, value_str TEXT, value_int INTEGER);";
	private static final String CREATE_TABLE_DICT = "CREATE TABLE " + TABLE_DICTINFO
			+ " (lang TEXT PRIMARY KEY, nb_entries INTEGER, table_name TEXT);";
	private static final String CREATE_TABLE_KANJI_ELEM = "CREATE TABLE "
			+ TABLE_KANJI_ELEM
			+ " (id INTEGER, num INTEGER, keb TEXT NOT NULL, pri TEXT, inf TEXT) CONSTRAINT kele_pk PRIMARY_KEY (id,num);";
	private static final String CREATE_TABLE_READING_ELEM = "CREATE TABLE "
			+ TABLE_READING_ELEM
			+ " (id INTEGER , num INTEGER, reb TEXT NOT NULL, pri TEXT, inf TEXT) CONSTRAINT rele_pk PRIMARY_KEY (id,num);";
	private static final String CREATE_TABLE_SENSE = "CREATE TABLE "
			+ TABLE_SENSE
			+ " (id INTEGER, num INTEGER, stagk TEXT, stagr TEXT, pos TEXT, xref TEXT) CONSTRAINT sense_pk PRIMARY_KEY (id,num);";
	private static final String CREATE_TABLE_GLOSS = "CREATE TABLE "
			+ TABLE_GLOSS
			+ " (id INTEGER, num INTEGER, gloss TEXT NOT NULL) CONSTRAINT gloss_pk PRIMARY_KEY (id,num);";

	private static final String CREATE_INDEX_ID = "CREATE INDEX idx_%1$s_id ON %1$s (id);";

	private SqlJetDb db;
	private JMDictReader jmDict;

	public MainJMDict(JMDictReader jmdict) {
		this.jmDict = jmdict;
	}

	public void createTables() throws SqlJetException {
		db.createTable(CREATE_TABLE_DICT);
		db.createTable(CREATE_TABLE_KANJI_ELEM);
		db.createTable(CREATE_TABLE_READING_ELEM);
		db.createTable(CREATE_TABLE_SENSE);
		db.createTable(CREATE_TABLE_DB_INFO);

		ArrayList<String> langs = this.jmDict.getLangs();
		for (String lang : langs) {
			lang = Common.fixLang(lang);
			db.createTable(String.format(CREATE_TABLE_GLOSS, lang));
		}
	}

	public void createDatabase(File dbFile) throws SqlJetException {
		long startTime = System.currentTimeMillis();

		this.db = new SqlJetDb(dbFile, true);
		this.db.open();
		this.db.getOptions().setAutovacuum(true);
		// utf16 ??? check phone speed and db size
		this.db.getOptions().setEncoding(SqlJetEncoding.UTF8);
		this.db.beginTransaction(SqlJetTransactionMode.WRITE);

		try {
			this.createTables();

			ISqlJetTable tableDB = this.db.getTable(TABLE_DBINFO);
			tableDB.insert(KEY_DB_VERSION, null, VALUE_DB_VERSION);
			tableDB.insert(KEY_DB_CREATION_DATE, VALUE_CREATION_DATE, null);

			this.writeEntriesSorted();

			// delete tables with no entries, and get others table info
			ArrayList<String> langs = this.jmDict.getLangs();
			ISqlJetTable tableInfo = this.db.getTable(TABLE_DICTINFO);
			for (String lang : langs) {
				lang = Common.fixLang(lang);
				String tableName = String.format(TABLE_GLOSS, lang);
				ISqlJetTable tableGloss = this.db.getTable(tableName);
				ISqlJetCursor cursor = tableGloss.open();

				String langNameEn = Common.getLangName(lang);
				if (!cursor.last()) {
					cursor.close();
					this.db.dropTable(tableName);
					System.out.printf("drop unuse %1$s table %2$s\n", langNameEn == null ? lang
							: langNameEn, tableName);
				} else {
					long count = cursor.getRowIndex();
					tableInfo.insert(lang, count, tableName);
					cursor.close();
					db.createIndex(String.format(String.format(CREATE_INDEX_ID, TABLE_GLOSS), lang));
					System.out.printf("%1$s: %2$d entries\n", langNameEn == null ? lang
							: langNameEn, count);
				}
			}

		} finally {
			this.db.commit();

			db.createIndex(String.format(CREATE_INDEX_ID, TABLE_KANJI_ELEM));
			db.createIndex(String.format(CREATE_INDEX_ID, TABLE_READING_ELEM));
			db.createIndex(String.format(CREATE_INDEX_ID, TABLE_SENSE));

			this.db.close();
		}

		long endTime = System.currentTimeMillis();
		System.out.println("create db time: " + ((endTime - startTime) / 1000) + "."
				+ ((endTime - startTime) % 1000) + "s");
	}

	private class EntryComparable implements Comparable<EntryComparable> {
		private String text;
		private int id;
		private int num;
		private String pri;
		private String inf;

		@Override
		public int compareTo(EntryComparable ec) {
			int d = text.length() - ec.text.length();
			d = d > 0 ? 1 : d;
			d = d < 0 ? -1 : d;
			if (d == 0)
				d = text.compareTo(ec.text);

			return d;
		}
	}

	private void writeEntriesSorted() throws SqlJetException {
		ArrayList<EntryComparable> readingElements;
		ArrayList<EntryComparable> kanjiElements;
		HashMap<String, ArrayList<EntryComparable>> mapSortedLang;

		mapSortedLang = new HashMap<String, ArrayList<EntryComparable>>(NUM_LANGS);
		readingElements = new ArrayList<MainJMDict.EntryComparable>(NUM_READING);
		kanjiElements = new ArrayList<MainJMDict.EntryComparable>(NUM_KANJI);

		ArrayList<Entry> entries = this.jmDict.getEntries();
		int numEntries = entries.size();

		for (int i = 0; i < numEntries; i++) {
			this.writeSense(entries.get(i), mapSortedLang, readingElements, kanjiElements);
		}
		this.db.commit();
		this.db.beginTransaction(SqlJetTransactionMode.WRITE);

		System.out.println("Sorting kanji elements...");
		Collections.sort(kanjiElements);
		int numKanjis = kanjiElements.size();
		ISqlJetTable table = this.db.getTable(TABLE_KANJI_ELEM);
		System.out.println("Writing kanji elements...");
		for (int i = 0; i < numKanjis; i++) {
			EntryComparable ec = kanjiElements.get(i);
			table.insert(ec.id, ec.num, ec.text, ec.pri, ec.inf);
		}
		this.db.commit();
		this.db.beginTransaction(SqlJetTransactionMode.WRITE);

		System.out.println("Sorting reading elements...");
		Collections.sort(readingElements);
		int numReading = readingElements.size();
		table = this.db.getTable(TABLE_READING_ELEM);
		System.out.println("Writing reading elements...");
		for (int i = 0; i < numReading; i++) {
			EntryComparable ec = readingElements.get(i);
			table.insert(ec.id, ec.num, ec.text, ec.pri, ec.inf);
		}
		this.db.commit();
		this.db.beginTransaction(SqlJetTransactionMode.WRITE);

		readingElements.clear();
		kanjiElements.clear();
		readingElements = null;
		kanjiElements = null;
		System.gc();

		Set<String> kSet = mapSortedLang.keySet();
		for (String lang : kSet) {
			ArrayList<EntryComparable> alEC = mapSortedLang.get(lang);
			String fixLang = Common.fixLang(lang);
			String fixLangName = Common.getLangName(fixLang);
			System.out.println("Sorting " + fixLangName + " glosses...");
			Collections.sort(alEC);
			int numGloss = alEC.size();
			System.out.println("Writing " + numGloss + " " + fixLangName + " glosses...");
			table = this.db.getTable(String.format(TABLE_GLOSS, fixLang));
			for (int i = 0; i < numGloss; i++) {
				EntryComparable sg = alEC.get(i);
				table.insert(sg.id, sg.num, sg.text);
			}
		}

		mapSortedLang.clear();
		mapSortedLang = null;
		this.db.commit();
		this.db.beginTransaction(SqlJetTransactionMode.WRITE);
		System.gc();

	}

	private void writeSense(Entry entry, HashMap<String, ArrayList<EntryComparable>> mapSortedLang,
			ArrayList<EntryComparable> readingElements, ArrayList<EntryComparable> kanjiElements)
			throws SqlJetException {
		// ISqlJetTable table = this.db.getTable(TABLE_KANJI_ELEM);
		int id = entry.sequence;

		for (int i = 0; i < entry.k_ele.size(); i++) {
			KanjiElement k_ele = entry.k_ele.get(i);
			String pri = this.listToString(k_ele.ke_pri);
			String inf = this.listToString(k_ele.ke_inf);
			// table.insert(id, i, k_ele.keb, pri, inf);
			EntryComparable ec = new EntryComparable();
			ec.id = id;
			ec.num = i;
			ec.text = k_ele.keb;
			ec.pri = pri;
			ec.inf = inf;

			kanjiElements.add(ec);
		}

		// table = this.db.getTable(TABLE_READING_ELEM);
		for (int i = 0; i < entry.r_ele.size(); i++) {
			ReadingElement r_ele = entry.r_ele.get(i);
			String pri = this.listToString(r_ele.re_pri);
			String inf = this.listToString(r_ele.re_inf);
			// table.insert(id, i, r_ele.reb, pri, inf);
			EntryComparable ec = new EntryComparable();
			ec.id = id;
			ec.num = i;
			ec.text = r_ele.reb;
			ec.pri = pri;
			ec.inf = inf;

			readingElements.add(ec);
		}

		ISqlJetTable table = this.db.getTable(TABLE_SENSE);
		for (int i = 0; i < entry.senses.size(); i++) {
			Sense sense = entry.senses.get(i);
			String stagk = this.listToString(sense.stagk);
			String stagr = this.listToString(sense.stagr);
			String pos = this.listToString(sense.pos);
			String xref = this.listToString(sense.xref);

			table.insert(id, i, stagk, stagr, pos, xref);

			HashMap<String, String> glossTextMap = new HashMap<String, String>();
			for (int j = 0; j < sense.glosses.size(); j++) {
				SenseGloss gloss = sense.glosses.get(j);

				String glossText = null;
				if (glossTextMap.containsKey(gloss.lang)) {
					glossText = glossTextMap.get(gloss.lang) + SLASH + gloss.text;
				} else {
					glossText = gloss.text;
				}
				glossTextMap.put(gloss.lang, glossText);
			}

			Set<String> kSet = glossTextMap.keySet();
			for (String lang : kSet) {
				// ISqlJetTable tableGloss =
				// this.db.getTable(String.format(TABLE_GLOSS,
				// Common.fixLang(lang)));
				// tableGloss.insert(id, i, glossTextMap.get(lang));
				EntryComparable ec = new EntryComparable();
				ec.text = glossTextMap.get(lang);
				ec.id = id;
				ec.num = i;

				ArrayList<EntryComparable> alEC = mapSortedLang.get(lang);
				if (null == alEC) {
					alEC = new ArrayList<MainJMDict.EntryComparable>(NUM_GLOSS_MAX);
					mapSortedLang.put(lang, alEC);
				}
				alEC.add(ec);
			}
		}

	}

	private String listToString(ArrayList<String> list) {
		if (null == list || 0 == list.size()) {
			return null;
		}

		String ret = list.get(0);
		for (int i = 1; i < list.size(); i++) {
			ret = ret + SLASH + list.get(i);
		}

		return ret;
	}

	private static void usage() {
		System.out.println("usage:");
		System.out.println("prog <JMdict_file> <db_file>");
		System.out.println("JMDict file direct download link: ftp://ftp.monash.edu.au/pub/nihongo/JMdict.gz");
	}

	public static void main(String[] args) {
		if (args.length != 2) {
			usage();
			return;
		}

		File jmDictFile = new File(args[0]);
		if (!(jmDictFile.isFile() && jmDictFile.canRead())) {
			System.err.printf("%s can't be read\n", args[0]);
			return;
		}

		File dbFile = new File(args[1]);
		if (dbFile.isDirectory()) {
			System.err.printf("%s is a directory\n", args[1]);
			return;
		}

		if (jmDictFile.equals(dbFile)) {
			System.err.printf("%s and %s is the same file\n", args[0], args[1]);
			return;
		}

		if (dbFile.isFile()) {
			System.out.printf("%s exists, deleting...\n", dbFile.getName());
			boolean del = dbFile.delete();
			if (!del) {
				System.err.printf("can't delete %s\n", args[1]);
				return;
			}
		}

		JMDictReader reader = new JMDictReader(jmDictFile);

		try {
			reader.read();
		} catch (XmlPullParserException xppe) {
			System.err.println("Error line " + reader.getLineNumber());
			xppe.printStackTrace();
			return;
		} catch (IOException ioe) {
			ioe.printStackTrace();
			return;
		}

		MainJMDict sqliteWriter = new MainJMDict(reader);
		try {
			sqliteWriter.createDatabase(dbFile);
		} catch (SqlJetException sje) {
			sje.printStackTrace();
		}
	}

}
