/**
 *  Copyright (C) 2012 Yvon TANGUY

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.vono.narau.kanjidb;

import java.io.File;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import java.util.Map.Entry;
import java.util.Set;

import org.tmatesoft.sqljet.core.SqlJetEncoding;
import org.tmatesoft.sqljet.core.SqlJetException;
import org.tmatesoft.sqljet.core.SqlJetTransactionMode;
import org.tmatesoft.sqljet.core.table.ISqlJetTable;
import org.tmatesoft.sqljet.core.table.SqlJetDb;
import org.vono.narau.kanjidb.Kanji.Element;

/**
 * TODO: put in common some part with JMdict (like TABLE_DBINFO)
 * 
 */
public class KanjiDBWriter {
	private static final String KEY_DB_VERSION = "version";
	private static final int VALUE_DB_VERSION = 2;
	private static final String KEY_DB_CREATION_DATE = "creation_date";
	private static final String VALUE_CREATION_DATE = String.format("%1$tY%1$tm%1$td",
			Calendar.getInstance());

	private static final String TABLE_DBINFO = "db_info";
	private static final String TABLE_RADPOS = "radpos"; // rapical + position
	private static final String TABLE_KANJI = "kanji";
	private static final String TABLE_MEANINGS = "meanings";
	/**
	 * can't used 'group' as the table name, it is a reserved word, the produced
	 * db can't be read.
	 */
	private static final String TABLE_GROUP = "kgroup";
	private static final String TABLE_PATH = "path";

	private static final String CREATE_TABLE = "CREATE TABLE ";

	private static final String CREATE_TABLE_DB_INFO = CREATE_TABLE + TABLE_DBINFO
			+ " (key TEXT PRIMARY KEY, value_str TEXT, value_int INTEGER);";

	/**
	 * The kanji table, merge the primary group. I took only the 2 first more
	 * used elements of the main group:<br/>
	 * <ul>
	 * <li>6404 kvg:element</li>
	 * <li>38 kvg:original</li>
	 * <li>3 kvg:partial</li>
	 * <li>1 kvg:phon</li>
	 * <li>4 kvg:position</li>
	 * <li>247 kvg:radical</li>
	 * <li>1 kvg:radicalForm</li>
	 * <li>31 kvg:variant</li>
	 * </ul>
	 * Unix cmd:
	 * 
	 * <pre>grep '^&lt;g id="kvg:....."' kanjivg.xml | sed 's/&lt;g
	 * id="kvg:....."//' | sed 's/="[^"]*"//g' | sed 's/&gt;//g' | sed 's/ /\n/g' |
	 * sort | uniq -c</pre>
	 */
	private static final String CREATE_TABLE_KANJI = CREATE_TABLE + TABLE_KANJI
			+ " (codepoint INTEGER PRIMARY KEY," + "element TEXT," + "radical INTEGER,"
			+ "grade INTEGER," + "frequency INTEGER," + "jlpt INTEGER,"
			+ "stroke_count INTEGER, reading_on TEXT, reading_kun TEXT);";

	private static final String CREATE_TABLE_MEANINGS = CREATE_TABLE + TABLE_MEANINGS
			+ " (codepoint INTEGER NOT NULL," + "lang TEXT NOT NULL," + "meanings TEXT NOT NULL"
			+ "); CONSTRAINT meanings_pk PRIMARY_KEY(codepoint, lang);";

	/**
	 * <ul>
	 * <li>5 kvg:radicalForm</li>
	 * <li>7 kvg:tradForm</li>
	 * <li>19 kvg:number</li>
	 * <li>468 kvg:partial</li>
	 * <li>1361 kvg:phon</li>
	 * <li>3425 kvg:part</li>
	 * <li>4487 kvg:original</li>
	 * <li>5892 kvg:variant</li>
	 * <li>35906 kvg:element</li>
	 * <li>7040 kvg:radical</li>
	 * <li>25228 kvg:position</li>
	 * </ul>
	 * 
	 * <pre>
	 * grep -E '<g id="kvg:.....-g[0-9]+"' kanjivg.xml | sed -E 's/.*<g id="kvg:.....-g[0-9]+"//'| sed 's/="[^"]*"//g' | sed 's/>//g' | sed 's/ /\n/g' | sort | uniq -c
	 * </pre>
	 */
	private static final String CREATE_TABLE_GROUP = CREATE_TABLE
			+ TABLE_GROUP
			+ " (codepoint INTEGER, id INTEGER, parent INTEGER NOT NULL, element TEXT, position INTEGER, radical INTEGER)"
			+ " CONSTRAINT group_pk PRIMARY_KEY (codepoint, id);";
	private static final String CREATE_TABLE_PATH = CREATE_TABLE
			+ TABLE_PATH
			+ " (codepoint INTEGER NOT NULL, parent INTEGER NOT NULL, type TEXT, draw TEXT NOT NULL)"
			+ " CONSTRAINT path_pk PRIMARY_KEY (codepoint, parent);";
	/**
	 * radical: 'r'+id ("bottom", "kamae", "kamae1", "kamae2", "left", "nyo",
	 * "right", "tare", "top")<br/>
	 * position: 'p'+id ("general", "nelson", "tradit")<br/>
	 */
	private static final String CREATE_TABLE_RADPOS = "CREATE TABLE " + TABLE_RADPOS
			+ " (id TEXT PRIMARY KEY, value TEXT NOT NULL);";

	private static final String CREATE_INDEX = "CREATE INDEX idx_%1$s_id ON %1$s (codepoint);";

	private SqlJetDb db;
	private ArrayList<Kanji> listKanji;

	public KanjiDBWriter(ArrayList<Kanji> kanjis) {
		this.listKanji = kanjis;
	}

	public void createDB(File dbFile) throws SqlJetException {
		long startTime = System.currentTimeMillis();
		System.out.println("Creating " + dbFile.getName() + "...");

		this.db = new SqlJetDb(dbFile, true);
		this.db.open();
		this.db.getOptions().setAutovacuum(true);
		this.db.getOptions().setEncoding(SqlJetEncoding.UTF8);
		this.db.beginTransaction(SqlJetTransactionMode.WRITE);

		try {
			this.createTables();

			ISqlJetTable table = this.db.getTable(TABLE_DBINFO);
			table.insert(KEY_DB_VERSION, null, VALUE_DB_VERSION);
			table.insert(KEY_DB_CREATION_DATE, VALUE_CREATION_DATE, null);

			// Map<Integer, Kanji> map = this.mapKanjis;
			// Set<Entry<Integer, Kanji>> set = this.mapKanjis.entrySet();
			ISqlJetTable tableKanji = this.db.getTable(TABLE_KANJI);
			ISqlJetTable tableGroup = this.db.getTable(TABLE_GROUP);
			ISqlJetTable tablePath = this.db.getTable(TABLE_PATH);
			ISqlJetTable tableMeanings = this.db.getTable(TABLE_MEANINGS);

			List<Kanji.Element> emptyList = new ArrayList<Kanji.Element>();
			ArrayList<Kanji> lList = this.listKanji;
			int count = lList.size();

			// for (Entry<Integer, Kanji> entry : set) {
			for (int i = 0; i < count; i++) {
				// Kanji kanji = entry.getValue();
				Kanji kanji = lList.get(i);
				int codepoint = kanji.codepoint;

				Integer radical = null;
				String element = null;

				Kanji.Group mainGroup = kanji.group;
				if (null != mainGroup) {
					radical = mainGroup.radical;
					element = mainGroup.element;
				}

				tableKanji.insert(codepoint, element, radical, kanji.grade, kanji.frequency,
						kanji.jlpt, kanji.strokeCount, kanji.readingsOn, kanji.readingsKun);

				List<Kanji.Element> elements = mainGroup == null ? emptyList : mainGroup.elements;
				int numElem = elements.size();
				for (int c = 0; c < numElem; c++) {
					writeElement(tableGroup, tablePath, codepoint, mainGroup, elements.get(c));
				}

				if (null != kanji.meanings) {
					Set<Entry<String, String>> meanings = kanji.meanings.entrySet();
					for (Entry<String, String> meaning : meanings) {
						tableMeanings.insert(codepoint, meaning.getKey(), meaning.getValue());
					}
				}
			}

			ISqlJetTable tableRadPos = this.db.getTable(TABLE_RADPOS);
			int len = Kanji.POSITIONS.length;
			for (int i = 0; i < len; i++) {
				tableRadPos.insert("p" + i, Kanji.POSITIONS[i]);
			}
			len = Kanji.RADICAL.length;
			for (int i = 0; i < len; i++) {
				tableRadPos.insert("r" + i, Kanji.RADICAL[i]);
			}

			this.db.commit();
			db.createView("create view kanjivg as "
					+ "select distinct kanji.codepoint as codepoint, kanji.jlpt as jlpt, kanji.grade as grade "
					+ "from kanji,path " + "where kanji.codepoint=path.codepoint");

		} finally {
			this.db.commit();

			db.createIndex(String.format(CREATE_INDEX, TABLE_KANJI));
			db.createIndex(String.format(CREATE_INDEX, TABLE_GROUP));
			db.createIndex(String.format(CREATE_INDEX, TABLE_PATH));
			db.createIndex(String.format(CREATE_INDEX, TABLE_MEANINGS));

			this.db.close();
		}

		long endTime = System.currentTimeMillis();
		System.out.println("create db time: " + ((endTime - startTime) / 1000) + "."
				+ ((endTime - startTime) % 1000) + "s");
	}

	private void writeElement(ISqlJetTable tableGroup, ISqlJetTable tablePath, int codepoint,
			Kanji.Group parent, Kanji.Element element) throws SqlJetException {
		if (element instanceof Kanji.Group) {
			Kanji.Group group = (Kanji.Group) element;
			tableGroup.insert(codepoint, group.id, parent.id, group.element, group.position,
					group.radical);

			List<Element> elements = group.elements;
			int numElem = elements.size();
			if (0 == numElem) {
				System.err.println("WARN: codepoint " + Integer.toHexString(codepoint) + " group "
						+ group.id + "with no sub element ");
			}
			for (int i = 0; i < numElem; i++) {
				writeElement(tableGroup, tablePath, codepoint, group, elements.get(i));
			}
		} else { // if (element instanceof Kanji.Path)
			Kanji.Path path = (Kanji.Path) element;
			tablePath.insert(codepoint, parent.id, path.type, path.draw);
		}
	}

	private void createTables() throws SqlJetException {
		this.db.createTable(CREATE_TABLE_DB_INFO);
		this.db.createTable(CREATE_TABLE_KANJI);
		this.db.createTable(CREATE_TABLE_GROUP);
		this.db.createTable(CREATE_TABLE_PATH);
		this.db.createTable(CREATE_TABLE_RADPOS);
		this.db.createTable(CREATE_TABLE_MEANINGS);
	}

}
